@@ -16,15 +16,16 @@
  * USA
  */
 
-#include <stdlib.h>
-#include <unistd.h>
-#include <pwd.h>
 #include <grp.h>
+#include <pwd.h>
+#include <stdlib.h>
 #include <sys/param.h>
+#include <unistd.h>
 
 #ifndef BSD
+#include <gshadow.h>
 #include <shadow.h>
-#endif // ifndef BSD
+#endif  // ifndef BSD
 
 #ifndef NSS_TEST_H
 #define NSS_TEST_H
@@ -33,7 +34,8 @@ const char *PASSWD_FILE = ".testdata/passwd.cache";
 const char *GROUP_FILE = ".testdata/group.cache";
 #ifndef BSD
 const char *SHADOW_FILE = ".testdata/shadow.cache";
-#endif // ifndef BSD
+const char *GSHADOW_FILE = ".testdata/gshadow.cache";
+#endif  // ifndef BSD
 
 extern enum nss_status _nss_cache_getpwent_r(struct passwd *result,
                                              char *buffer, size_t buflen,
@@ -59,7 +61,12 @@ extern enum nss_status _nss_cache_getspnam_r(const char *name,
                                              size_t buflen, int *errnop);
 extern enum nss_status _nss_cache_getspent_r(struct spwd *result, char *buffer,
                                              size_t buflen, int *errnop);
+extern enum nss_status _nss_cache_getsgnam_r(const char *name,
+                                             struct sgrp *result, char *buffer,
+                                             size_t buflen, int *errnop);
+extern enum nss_status _nss_cache_getsgent_r(struct sgrp *result, char *buffer,
+                                             size_t buflen, int *errnop);
 extern char *_nss_cache_setpwent_path(const char *path);
-#endif // ifndef BSD
+#endif  // ifndef BSD
 
 #endif /* NSS_TEST_H */
